/* povwrite string generator: strings composed out of 3d letters
 * for POVRAY 2.0 
 *
 * Copyright by Heiko Eissfeldt and Michael Bischoff
 *
 * GNU Public License 2 applies
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * this version: 30-aug-1994 by Heiko Eissfeldt and Michael Bischoff
 *
 * this is the ISO 8815-1 version. Extended ascii systems need 
 * a character conversion of this source file (use eg recode).
 *
 * to compile, type: "gcc -s -N -O2 -o pw pw.c -lm"
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef M_PI
#define M_PI	3.14159265358979
#endif

#define RAD(a)	        ((a) / 180.0 * M_PI)
#define DEGREE(a)	(180.0 / M_PI * (a))
#define TRUE		1

double cx = 0.0, cy = 0.0, cz = 0.0;
double dx = 1.0, dy = 0.0, dz = 0.0;

#define FAR_AWAY	1e37
double xmax = -FAR_AWAY, xmin = FAR_AWAY;
double ymax = -FAR_AWAY, ymin = FAR_AWAY;
double zmax = -FAR_AWAY, zmin = FAR_AWAY;

#define SETMAX(x, y, z)	{		\
	if ((x) > xmax) xmax = (x);	\
	if ((y) > ymax) ymax = (y);	\
	if ((z) > zmax) zmax = (z);	\
	if ((x) < xmin) xmin = (x);	\
	if ((y) < ymin) ymin = (y);	\
	if ((z) < zmin) zmin = (z);	\
}

/* tunable parameters: */
double Rad = 0.3;	/* 1st toroid radius for 'o' */
double rad = 0.07;	/* 2nd toroid radius (thickness) */
double botsize = 0.4;	/* how much should ie 'g' or 'y' dive in? */
double topsize = 0.4;	/* how much should ie 'h' or come out? */
double npart = 0.80;	/* radius of "n"-arc and "h", "u"-arc, counted in o-radius */
double mpart = 0.60;	/* same for "m" */
double fpart = 0.40;	/* upper arc of 'f' */
double szpart = 0.60;	/* upper arc of '' */
double kwidth = 1.2;	/* width of k in Rad */
double abar = 0.35;	/* height of A-bar relative to A height */
double cangle = 30.0;	/* angle of missing toroid segment in 'c' */
double eangle = 40.0;	/* angle of missing toroid segment in 'e' */
double sangle = 50.0;	/* angle of missing toroid segment in 'c' */
double umlautdx = 0.07; /* points of umlauts are that far apart */
double braceangle = 40.0; /* halved angle for braces '(' ')'*/
double gap = 0.07;	/* gap in the middle of '|' */
/* end of tunable parameters */

#define NO_BALLS	0
#define BOT_BALL	1
#define LEFT_BALL	1
#define FIRST_BALL	1
#define TOP_BALL	2
#define RIGHT_BALL	2
#define END_BALL	2
#define SECOND_BALL	2
#define BOTH_BALLS	3

#define TOP	(Rad + Rad + topsize)	/* total height */

/* design is bottom-up (each function calls only functions above it) */

/* for toroid, see documentation, r1 = 2 * (Rad+rad), r0 = 2 * (Rad - rad) */

void declare_toroid(void)
{   double r0, r1, plus, minus;
    r1 = rad * rad;
    r0 = Rad * Rad;
    plus =  r0 + r1;
    minus = r0 - r1;
    printf("#declare Toroid = quartic { <\n");
    printf("    1.0,   0.0,   0.0,   0.0,   2.0,\n");
    printf("    0.0,   0.0,   2.0,   0.0,   %f,\n", -2.0 * plus);
    printf("    0.0,   0.0,   0.0,   0.0,   0.0,\n");
    printf("    0.0,   0.0,   0.0,   0.0,   0.0,\n");
    printf("    1.0,   0.0,   0.0,   2.0,   0.0,\n");
    printf("    %f,    0.0,   0.0,   0.0,   0.0,\n", -2.0 * plus);
    printf("    1.0,   0.0,   %f,    0.0,   %f\n", 2.0 * minus, minus * minus);
    printf("    > }\n");
}

void uv_toroid(void)
{   double r0, r1, plus, minus, ystretch;
    r1 = rad * rad;
    r0 = Rad * Rad;
    plus =  r0 + r1;
    minus = r0 - r1;
    ystretch = (Rad-rad/4.0)/(TOP/2.0);
    ystretch *= ystretch; 
    printf("#declare uvToroid = quartic { <\n");
    /*     a0  (x^4)     = a20 (y^4)     = a30 (z^4)     = 1,
           a4  (x^2 y^2) = a7  (x^2 z^2) = a23 (y^2 z^2) = 2,
           a9  (x^2)     = a25 (y^2)     = -2 (r0^2 + r1^2)
           a32 (z^2)     = 2 (r0^2 - r1^2)
           a34 (1)       = 2 (r0^2 - r1^2)^2      */
/*              x^4    0      0      0      x^2 y^2   */
    printf("    1.0,   0.0,   0.0,   0.0,   %f,\n", 2.0 * ystretch);
/*              0      0    x^2z^2   0      x^2       */
    printf("    0.0,   0.0,   2.0,   0.0,   %f,\n", -2.0 * plus);
    printf("    0.0,   0.0,   0.0,   0.0,   0.0,\n");
    printf("    0.0,   0.0,   0.0,   0.0,   0.0,\n");
/*              y^4    0      0    y^2z^2   0         */
/*  printf("    %f,    0.0,   0.0,   %f,    0.0,\n",ystretch*ystretch*1.1, 2.0 * ystretch); */
    printf("    %f,    0.0,   0.0,   %f,    0.0,\n",ystretch*ystretch*1.13, 2.0 * ystretch);
/*              y^2    0      0      0      0         */
/*  printf("    %f,    0.0,   0.0,   0.0,   0.0,\n", -2.0 * plus * ystretch / 1.0008);*/
    printf("    %f,    0.0,   0.0,   0.0,   0.0,\n", -2.0 * plus * ystretch / 1.004);
/*              z^4    0      z^2    0      1         */
    printf("    1.0,   0.0,   %f,    0.0,   %f\n", 2.0 * minus, minus * minus);
    printf("    > }\n");
}

void r_Toroid(double R)
{   double r0, r1, plus, minus;
    r1 = rad * rad;
    r0 = R * R;
    plus =  r0 + r1;
    minus = r0 - r1;
    printf("\t    quartic { <\n");
    printf("\t\t1.0,   0.0,   0.0,   0.0,   2.0,\n");
    printf("\t\t0.0,   0.0,   2.0,   0.0,   %f,\n", -2.0 * plus);
    printf("\t\t0.0,   0.0,   0.0,   0.0,   0.0,\n");
    printf("\t\t0.0,   0.0,   0.0,   0.0,   0.0,\n");
    printf("\t\t1.0,   0.0,   0.0,   2.0,   0.0,\n");
    printf("\t\t%f,    0.0,   0.0,   0.0,   0.0,\n", -2.0 * plus);
    printf("\t\t1.0,   0.0,   %f,    0.0,   %f\n", 2.0 * minus, minus * minus);
    printf("\t    > }\n");
}

void ball(double x, double y, int flag)
{   SETMAX(cx+x, cy+y, cz);
    if (flag)
	printf("\tsphere { <%f, %f, %f>, %f }\n", cx+x, cy+y, cz, rad);
}

void flat_ball(double x, double y, int flag, double shrink)
{   SETMAX(cx+x, cy+y, cz);
    if (flag)
	printf("\tsphere { <%f, %f, %f>, %f\n\t   scale <1.0, 1.0, %f> }\n",
	       cx+x, cy+y, cz, rad, shrink);
}

void vertical_bar(double x, double y0, double y1, int flags)
{   ball(x, y1, flags & TOP_BALL);
    ball(x, y0, flags & BOT_BALL);
    printf("\tintersection {\n\t    object { Cylinder_Y\n\t\tscale <%f, %f, %f>"
	   "\n\t\ttranslate <%f, 0, %f>\n\t    }\n", rad, rad, rad, cx+x, cz);
    printf("\t    plane { <0,  1, 0>, %f }\n\t    plane { <0, -1, 0>, %f }\n\t}\n",
	cy + y1, -(cy + y0));
}

void horizontal_bar(double x0, double x1, double y, int flags)
{   ball(x0, y, flags & LEFT_BALL);
    ball(x1, y, flags & RIGHT_BALL);
    printf("\tintersection {\n\t    object { Cylinder_X\n\t\tscale <%f, %f, %f>"
	   "\n\t\ttranslate <0, %f, %f>\n\t    }\n", rad, rad, rad, cy+y, cz);
    printf("\t    plane { < 1, 0, 0>, %f }\n\t    plane { <-1, 0, 0>, %f }\n\t}\n",
	cx + x1, -(cx + x0));
}

void smallvert(double x)
{   vertical_bar(x, 0.0, 2.0 * Rad, BOTH_BALLS);
}

void topvert(double x)
{   vertical_bar(x, 0.0, 2.0 * Rad + topsize, BOTH_BALLS);
}

void botvert(double x)
{   vertical_bar(x, -botsize, 2.0 * Rad, BOTH_BALLS);
}

void tophor( double x )
{   horizontal_bar(0.0, x, TOP, BOTH_BALLS);
}

void std_toroid(void)
{   printf("\tobject { Toroid\n");
    printf("\t    translate <%f, %f, %f>\n", cx + Rad, cy + Rad, cz);
    printf("\t}\n");
    SETMAX(cx+Rad - Rad, cy+Rad, cz);
    SETMAX(cx+Rad + Rad, cy+Rad, cz);
    SETMAX(cx+Rad, cy - Rad+Rad, cz);
    SETMAX(cx+Rad, cy + Rad+Rad, cz);
}

void std_toroid2( double x, double y )
{   printf("\tobject { Toroid\n");
    printf("\t    translate <%f, %f, %f>\n", cx + x + Rad, cy + y + Rad, cz);
    printf("\t}\n");
    SETMAX(cx+x+Rad - Rad, cy+y+Rad, cz);
    SETMAX(cx+x+Rad + Rad, cy+y+Rad, cz);
    SETMAX(cx+x+Rad, cy - Rad+y+Rad, cz);
    SETMAX(cx+x+Rad, cy + Rad+y+Rad, cz);
}

void egg_toroid(void)
{   printf("\tobject { uvToroid\n");
    printf("\t    translate <%f, %f, %f>\n", cx + Rad, cy + TOP/2.0, cz);
    printf("\t}\n");
    SETMAX(cx+Rad - Rad, cy+Rad, cz);
    SETMAX(cx+Rad + Rad, cy+Rad, cz);
    SETMAX(cx+Rad, cy - Rad+Rad-Rad, cz);
    SETMAX(cx+Rad, cy + Rad+Rad+Rad, cz);
}

void toroid(double x, double y, double r, double angle0, double angle1, int flags)
{   double b1x, b1y, b2x, b2y;

    angle0 = fmod(angle0 + 3600.0, 360.0);
    angle1 = fmod(angle1 + 3600.0, 360.0);
    if (angle1 < angle0)
	angle1 += 360.0;
    if (angle1 == angle0) { /* do a full circle */
	if (r == 1.0)
	    printf("\t    object { Toroid\n");
	else {
	    printf("\t    object { \n");
	    r_Toroid(r * Rad);
	}
	printf("\t\t    translate <%f, %f, %f>\n",
	    cx + x, cy + y, cz);
	printf("\t}\n");
	SETMAX(cx+x - r*Rad, cy+y, cz);
	SETMAX(cx+x + r*Rad, cy+y, cz);
	SETMAX(cx+x, cy - r*Rad+y, cz);
	SETMAX(cx+x, cy + r*Rad+y, cz);
    } else {
	double r0, r1;
	printf("\tintersection {\n");
	if (r == 1.0)
	     printf("\t    object { Toroid }\n");
	else
	     r_Toroid(r * Rad);

	r0 = RAD(angle0);
	r1 = RAD(angle1);
	b1x = x + r * Rad * cos(r0);
	b1y = y + r * Rad * sin(r0);
	b2x = x + r * Rad * cos(r1);
	b2y = y + r * Rad * sin(r1);
	if (angle0 + 180.0 >= angle1) {		/* at most half a toroid */
	    printf("\t    plane { <%f, %f, 0>, 0 }\n", sin(r0), -cos(r0));
	    if (angle0 + 180.0 > angle1)		/* less than half a toroid */
		printf("\t    plane { <%f, %f, 0>, 0 }\n", -sin(r1), cos(r1));
	} else {				/* more than half a toroid */
	    printf("\t    union {\n");
	    printf("\t\tplane { <%f, %f, 0>, 0 }\n",  sin(r0), -cos(r0));
	    printf("\t\tplane { <%f, %f, 0>, 0 }\n", -sin(r1),  cos(r1));
	    printf("\t    }\n");
	}

	printf("\t    translate <%f, %f, %f>\n",
	    cx + x, cy + y, cz);
	printf("\t}\n");
	ball(b1x, b1y, flags & FIRST_BALL);
	ball(b2x, b2y, flags & END_BALL);

	/* compute bounding points */
	if (angle1 > 360.0) SETMAX(cx+x+r*Rad, cy+y, cz);
	angle0 = fmod(angle0 + 270.0, 360.0);
	angle1 = fmod(angle1 + 270.0, 360.0);
	if (angle1 < angle0) angle1 += 360.0;
	if (angle1 > 360.0) SETMAX(cx+x, cy+r*Rad+y, cz);
	angle0 = fmod(angle0 + 270.0, 360.0);
	angle1 = fmod(angle1 + 270.0, 360.0);
	if (angle1 < angle0) angle1 += 360.0;
	if (angle1 > 360.0) SETMAX(cx+x-r*Rad, cy+y, cz);
	angle0 = fmod(angle0 + 270.0, 360.0);
	angle1 = fmod(angle1 + 270.0, 360.0);
	if (angle1 < angle0) angle1 += 360.0;
	if (angle1 > 360.0) SETMAX(cx+x, cy-r*Rad+y, cz);
    }
}

void ruv_Toroid(double R)
{
  r_Toroid(R);
}

void oval_toroid(double x, double y, double r, double angle0, double angle1, int flags)
{   double b1x, b1y, b2x, b2y;

    angle0 = fmod(angle0 + 3600.0, 360.0);
    angle1 = fmod(angle1 + 3600.0, 360.0);
    if (angle1 < angle0)
	angle1 += 360.0;
    if (!(angle1 - angle0)) { /* do a full circle */
	if (r == 1.0)
	    printf("\t    object { uvToroid }\n");
	else
	    ruv_Toroid(r * Rad);
	printf("\t    translate <%f, %f, %f>\n",
	    cx + x, cy + y, cz);
	SETMAX(cx+x - r*Rad, cy+y, cz);
	SETMAX(cx+x + r*Rad, cy+y, cz);
	SETMAX(cx+x, cy - r*(TOP/2.0)+y, cz);
	SETMAX(cx+x, cy + r*(TOP/2.0)+y, cz);
    } else {
	double r0, r1;
	double r0i, r1i;
	double shift0, shift1;
	printf("\tintersection {\n");
	if (r == 1.0)
	     printf("\t    object { uvToroid }\n");
	else
	     ruv_Toroid(r * Rad);

	r0 = RAD(angle0);
	r1 = RAD(angle1);
	b1x = x + r * Rad * cos(r0);
	b1y = y + r * (TOP/2.0) * sin(r0);
	b2x = x + r * Rad * cos(r1);
	b2y = y + r * (TOP/2.0) * sin(r1);

	if (r0 < 0.00001 && r0 > -0.00001)
	    r0i = RAD(0) + (cos(r0) > 0 ? 0 : RAD(180));
	else
	    r0i = atan2(r * (TOP/2.0) * sin(r0) - (TOP/2.0 - Rad),
			r * Rad * cos(r0));
	if (r1 < 0.00001 && r1 > -0.00001)
	    r1i = RAD(0) + (cos(r1) > 0 ? 0 : RAD(180));
	else
	    r1i = atan2(r * (TOP/2.0) * sin(r1) + (TOP/2.0 - Rad),
			r * Rad * cos(r1));
			
	shift0 = -cos(r0i) * (TOP/2.0 - Rad);
	shift1 = -cos(r1i) * (TOP/2.0 - Rad);

	printf("\t    union {\n");
	printf("\t\tplane { <%f, %f, 0>, %f }\n", sin(r0i),-cos(r0i), shift0);
	printf("\t\tplane { <%f, %f, 0>, %f }\n", -sin(r1i),cos(r1i), shift1);
	printf("\t\tplane { <%f, 0, 0>, 0 }\n",angle1 <= 360.0 ? 1.0 : -1.0);
	printf("\t    }\n");

	printf("\t    translate <%f, %f, %f>\n",
	    cx + x, cy + y, cz);
	printf("\t}\n");
	/* ideally, we would compute the thickness at the given point. */
	/* The current value seems to be practical for 'C' */
	flat_ball(b1x, b1y, flags & FIRST_BALL, 0.98);
	flat_ball(b2x, b2y, flags & END_BALL, 0.98);

	/* compute bounding points */
	if (angle1 > 360.0) SETMAX(cx+x+r*Rad, cy+y, cz);
	angle0 = fmod(angle0 + 270.0, 360.0);
	angle1 = fmod(angle1 + 270.0, 360.0);
	if (angle1 < angle0) angle1 += 360.0;
	if (angle1 > 360.0) SETMAX(cx+x, cy+r*(TOP/2.0)+y, cz);
	angle0 = fmod(angle0 + 270.0, 360.0);
	angle1 = fmod(angle1 + 270.0, 360.0);
	if (angle1 < angle0) angle1 += 360.0;
	if (angle1 > 360.0) SETMAX(cx+x-r*Rad, cy+y, cz);
	angle0 = fmod(angle0 + 270.0, 360.0);
	angle1 = fmod(angle1 + 270.0, 360.0);
	if (angle1 < angle0) angle1 += 360.0;
	if (angle1 > 360.0) SETMAX(cx+x, cy-r*(TOP/2.0)+y, cz);
    }
}

void cylinder(double x1, double y1, double x2, double y2, int flags)
{   double x, y, len, alpha;
    alpha = atan((x2-x1) / (y2-y1));
    len = sqrt((x2-x1) * (x2-x1) + (y2-y1) * (y2-y1));
    x = x1 * cos(alpha) - y1 * sin(alpha);
    y = x1 * sin(alpha) + y1 * cos(alpha);

    ball(x2, y2, flags & TOP_BALL);
    ball(x1, y1, flags & BOT_BALL);
    printf("\tintersection {\n\t    object { Cylinder_Y\n\t\tscale <%f, %f, %f>\n"
	   "\t\ttranslate <%f, 0, 0>\n\t    }\n", rad, rad, rad, x);
    printf("\t    plane { <0,  1, 0>, %f }\n\t    plane { <0, -1, 0>, %f }\n",
	y + len, -y);
    printf("\t    rotate <0, 0, %f>\n", -DEGREE(alpha));
    printf("\t    translate <%f, %f, %f>\n\t}\n", cx, cy, cz);
}

/* advance the current position 2.0 * (D * Rad + d * rad) * (dx,dy,dz) */
/* by changing (dx,dy,dz), various forms can be achieved */
/* (even reverse or top/down writing) */

void advance(double D, double d)
{
    d = 2.0 * D * Rad + 2.0 * d * rad;
    cx += d * dx;
    cy += d * dy;
    cz += d * dz;
}


void writeletter(int c)
{   if (c == ' ')
	printf("\t/* a space */\n");
    else
	printf("\t/* the character %c */\n", c);
    switch (c) {
    /* missing : ~&{} */
    case ',':   cylinder(0.0, -0.5*Rad, Rad/4.0, 0.0, BOTH_BALLS);
		advance(0.125, 1.5);
		break;
    case '\'':  cylinder(0.0, TOP - Rad/2.0, Rad/4.0, TOP, BOTH_BALLS);
		advance(0.125, 1.5);
		break;
    case '`':   cylinder(Rad/4.0, TOP - Rad/2.0, 0.0, TOP, BOTH_BALLS);
		advance(0.125, 1.5);
		break;
    case '"':   cylinder(0.0, TOP - Rad/2.0, 0.0, TOP, BOTH_BALLS);
		cylinder(Rad, TOP - Rad/2.0, Rad, TOP, BOTH_BALLS);
		advance(0.5, 1.5);
		break;
    case '^':   cylinder(0.0, TOP - Rad/2.0, Rad/2.0, TOP, BOTH_BALLS);
		cylinder(Rad/1.0, TOP - Rad/2.0, Rad/2.0, TOP, BOTH_BALLS);
		advance(0.5, 1.5);
		break;
    case '%':	cylinder(0.0, 0.0, 2.0 * npart * Rad, TOP, BOTH_BALLS);
#if	0
		ball(0.0, TOP, TRUE);
		ball(2.0 * npart * Rad, 0.0, TRUE);
#else
	        toroid(TOP/8.0, 7.0*TOP/8.0, TOP/8.0/Rad,0,0,NO_BALLS);
	        toroid(2.0*npart*Rad-TOP/8.0,TOP/8.0,TOP/8.0/Rad,0,0,NO_BALLS);
#endif
		advance(npart, 1.5);
		break;
    case '.':	ball(0.0, 0.0, TRUE);
		advance(0.0, 1.5);
		break;
    case ';':   cylinder(0.0, -0.5*Rad, Rad/4.0, 0.0, BOTH_BALLS);
                ball(Rad/4.0, Rad, TRUE);
		advance(0.125, 1.5);
		break;
    case ':':	ball(0.0, Rad+1.5*rad, TRUE);
                ball(0.0, Rad-1.5*rad, TRUE);
		advance(0.125, 1.5);
		break;
    case '=':	horizontal_bar(0.0, 2.0 * Rad * npart, TOP/2.0+1.5*rad, BOTH_BALLS);
		horizontal_bar(0.0, 2.0 * Rad * npart, TOP/2.0-1.5*rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '+':	vertical_bar(Rad * npart, TOP/4.0, 3.0*TOP/4.0, BOTH_BALLS);
		horizontal_bar(0.0, 2.0 * Rad * npart, TOP/2.0, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '-':	horizontal_bar(0.0, 2.0 * Rad * npart, TOP/2.0, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '_':	horizontal_bar(-1.5*rad, 2.0 * Rad * npart + 1.5 * rad, 0.0, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '#':	horizontal_bar(0.0, 2.0 * Rad , 2.0*TOP/3.0, BOTH_BALLS);
		horizontal_bar(0.0, 2.0 * Rad , TOP/3.0, BOTH_BALLS);
		vertical_bar(2.0*Rad/3.0 , 0.0, TOP, BOTH_BALLS);
		vertical_bar(4.0*Rad/3.0 , 0.0, TOP, BOTH_BALLS);
		advance(1.0, 1.5);
		break;
    case '*':	cylinder(npart/2.0 * Rad, TOP/2.0 - Rad*sin(RAD(60.0)), 3.0/2.0 * npart * Rad, TOP/2.0 + Rad*sin(RAD(60.0)), BOTH_BALLS);
		cylinder(3.0/2.0 * npart * Rad, TOP/2.0 - Rad*sin(RAD(60.0)), npart/2.0 * Rad, TOP/2.0 + Rad*sin(RAD(60.0)), BOTH_BALLS);
               horizontal_bar(0.0, 2.0 * Rad * npart, TOP/2.0, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '<':	cylinder(0.0, TOP/2.0, 2.0 * npart * Rad, TOP/2.0 + Rad*sin(RAD(60.0)), BOTH_BALLS);
		cylinder(2.0 * npart * Rad, TOP/2.0 - Rad*sin(RAD(60.0)), 0.0, TOP/2.0 , BOT_BALL);
		advance(npart, 1.5);
		break;
    case '>':	cylinder(2.0 * npart * Rad, TOP/2.0, 0.0, TOP/2.0 + Rad*sin(RAD(60.0)), BOTH_BALLS);
		cylinder(0.0, TOP/2.0 - Rad*sin(RAD(60.0)), 2.0 * npart * Rad, TOP/2.0 , BOT_BALL);
		advance(npart, 1.5);
		break;
    case '\\':	cylinder(2.0 * npart * Rad, 0.0, 0.0, TOP, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '/':	cylinder(0.0, 0.0, 2.0 * npart * Rad, TOP, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case '[':	vertical_bar(0.0, 0.0, TOP, BOTH_BALLS);
		horizontal_bar(0.0, npart * Rad, 0.0, RIGHT_BALL);
		horizontal_bar(0.0, npart * Rad, TOP, RIGHT_BALL);
		advance(npart/2.0, 1.5);
		break;
    case ']':	vertical_bar(npart * Rad, 0.0, TOP, BOTH_BALLS);
		horizontal_bar(0.0, npart * Rad, 0.0, LEFT_BALL);
		horizontal_bar(0.0, npart * Rad, TOP, LEFT_BALL);
		advance(npart/2.0, 1.5);
		break;
    case '(':	{  
                  toroid(TOP/2.0/sin(RAD(braceangle)), TOP/2.0, 
                        TOP/2.0/Rad/sin(RAD(braceangle)), 
                     180.0 - braceangle, 180.0 + braceangle, BOTH_BALLS);
		   advance((TOP/2.0/sin(RAD(braceangle))*
                           (1.0-cos(RAD(braceangle))))/2.0/Rad, 1.5);
		break;
               }
    case ')':	{  
                  toroid(TOP/2.0/sin(RAD(braceangle))*
                           (1.0-cos(RAD(braceangle)))-
                        TOP/2.0/sin(RAD(braceangle)), TOP/2.0, 
                        TOP/2.0/Rad/sin(RAD(braceangle)), 
                     - braceangle, + braceangle, BOTH_BALLS);
		   advance((TOP/2.0/sin(RAD(braceangle))*
                           (1.0-cos(RAD(braceangle))))/2.0/Rad, 1.5);
		break;
               }
    case '|':	vertical_bar(0.0, TOP/2.0 + gap, TOP, BOTH_BALLS);
		vertical_bar(0.0, 0.0, TOP/2.0 - gap, BOTH_BALLS);
		advance(0.0, 1.5);
		break;
    case '!':	ball(0.0, 0.0, TRUE);
		vertical_bar(0.0, 3.0 * rad, TOP, BOTH_BALLS);
		advance(0.0, 1.5);
		break;
    case '?':	{   double x, r, cwi, beta, l, l2, tophalv = (TOP-3.0*rad)/2.0;
		    r = 0.35;	/* factor for small radius */
		    cwi = RAD(sangle);
		   /* distance between center and center of small circle in rad */
		    l = (1.0 - r) * tophalv;
		   /* x value for center in rad */
		   /* x = tophalv * (cos(cwi) * (1 - r) + r) */
		    x = cos(cwi) * l + r * tophalv;
		    toroid(x, tophalv+3.0*rad, tophalv/Rad, sangle, 180.0-sangle, SECOND_BALL);
		    toroid(x, tophalv+3.0*rad, tophalv/Rad, 180.0+sangle, -sangle, RIGHT_BALL);
		    beta = asin(r / (1.0 - r));	/* cwi - beta is the slope of center line segment */
		    if (beta <= 0.0 || beta >= cwi) {
			fprintf(stderr, "internal parameters incorrect\n");
			exit(1);
		    }
		   /* angle = 90 + beta */
		   toroid(x - l * cos(cwi), tophalv - l * sin(cwi) +3.0*rad,
			tophalv * r/Rad, 90.0 + sangle - DEGREE(beta), 180.0 + sangle, NO_BALLS);
		    toroid(x + l * cos(cwi), tophalv + l * sin(cwi) +3.0*rad,
			tophalv * r/Rad, sangle - 90.0 - DEGREE(beta), sangle, NO_BALLS);
		    l2 = sqrt(1.0 - 2.0 * r /* + r*r - r*r */) * tophalv;
		    cylinder(x - l2 * cos(cwi-beta),
                            tophalv - l2 * sin(cwi-beta) +3.0*rad,
			     x + l2 * cos(cwi-beta), 
                            tophalv + l2 * sin(cwi-beta) +3.0*rad, NO_BALLS);
                   ball(x, 0.0, TRUE);
		    advance(x / Rad, 1.5);
		}
		break;
    case '@':
#if	0
	        toroid(TOP/2.0, TOP/2.0, TOP/4.0/Rad,0,0,NO_BALLS);
		toroid(TOP/2.0, TOP/2.0, (TOP/2.0)/Rad, 90.0, -50.0, SECOND_BALL);
		toroid(TOP/2.0, (5.0*TOP)/8.0, (3.0*TOP/8.0)/Rad, -90.0, 90.0, NO_BALLS);
		advance(TOP/2.0/Rad, 1.5);
#else
                toroid(TOP/2.0, TOP/2.0, TOP/6.0/Rad,0,0,NO_BALLS);
		toroid(TOP/2.0, TOP/2.0, (TOP/2.0)/Rad, 0.0, -60.0, SECOND_BALL);
		toroid((5.0*TOP)/6.0, TOP/2.0, TOP/6.0/Rad, 180.0, 0.0, NO_BALLS);
	        vertical_bar(2.0*TOP/3.0, TOP/2.0, 2.0*TOP/3.0,TOP_BALL);
		advance(TOP/2.0/Rad, 1.5);
#endif
	        break;
    case 'A':	cylinder(0.0, 0.0, Rad, TOP, BOTH_BALLS);
		cylinder(2.0 * Rad, 0.0, Rad, TOP, BOT_BALL);
		horizontal_bar(abar * Rad, (2.0 - abar) * Rad, TOP*abar, NO_BALLS);
		advance(1.0, 1.5);
		break;
    case 'B':	vertical_bar(0.0, 0.0, TOP, BOTH_BALLS);
		horizontal_bar(0.0, npart * Rad, 0.0, NO_BALLS);
		horizontal_bar(0.0, npart * Rad, TOP, NO_BALLS);
		horizontal_bar(0.0, npart * Rad, TOP/2.0, NO_BALLS);
		toroid(npart * Rad, 3.0*TOP/4.0, (TOP/4.0)/Rad, -90.0, 90.0, NO_BALLS);
		toroid(npart * Rad, 1.0*TOP/4.0, (TOP/4.0)/Rad, -90.0, 90.0, NO_BALLS);
		advance(npart/2.0 + TOP/8.0/Rad, 1.5);
		break;
    case 'C':	oval_toroid(Rad, TOP/2.0, 1.0, cangle, -cangle, BOTH_BALLS);
		advance(1.0, 1.5);
		break;

    /* this one could be better */
    case 'D':	vertical_bar(0.0, 0.0, TOP, BOTH_BALLS);
		horizontal_bar(0.0, npart /2.0 * Rad, TOP, NO_BALLS);
		horizontal_bar(0.0, npart /2.0 * Rad, 0.0, NO_BALLS);
#if	0
		oval_toroid(npart /2.0 * Rad, TOP/2.0, 1.0, -90.0, 90.0, NO_BALLS);
#else
		toroid(npart/2.0 * Rad, TOP/2.0, (TOP/2.0)/Rad, -90.0, 90.0, NO_BALLS);
#endif
		advance(npart/2.0 + TOP/4.0/Rad, 1.5);
		break;
    case 'E':	horizontal_bar(0.0, 2.0 * npart * Rad, 0.0, RIGHT_BALL);
	        /* FALLTHROUGH */
    case 'F':	topvert(0.0);
		horizontal_bar(0.0, 2.0 * npart * Rad, TOP, RIGHT_BALL);
		horizontal_bar(0.0, 1.5 * npart * Rad, TOP/2.0, RIGHT_BALL);
		advance(npart, 1.5);
		break;
    case 'G':	oval_toroid(Rad, TOP/2.0, 1.0, cangle, 0.0, LEFT_BALL);
		horizontal_bar(Rad,2.0*Rad, TOP/2.0, BOTH_BALLS);
                vertical_bar(2.0*Rad, 0.0, TOP/2.0, BOT_BALL);
		advance(1.0, 1.5);
		break;
    case 'H':	topvert(0.0);
		topvert(2.0 * Rad * npart);
		horizontal_bar(0.0, 2.0 * Rad * npart, TOP/2.0, NO_BALLS);
		advance(npart, 1.5);
		break;
    case 'I':	topvert(0.0);
		advance(0.0, 1.5);
		break;
    case 'J':   advance(npart/2.0 * (1.0 + cos(RAD(sangle))), 0.0);
                vertical_bar(0.0, npart * Rad, TOP, BOTH_BALLS);
		toroid(-npart * Rad, npart * Rad, npart, 180.0 + sangle, 0.0, FIRST_BALL);
		advance(0.0, 1.5);
		break;
    case 'K':	topvert(0.0);
		cylinder(0.0, TOP/3.0, 2.0 * npart * Rad, TOP, TOP_BALL);
		cylinder(2.0 * npart * Rad, 0.0, Rad * npart / 2.0, TOP / 2.0, BOT_BALL);
		advance(npart, 1.5);
		break;
    case 'L':	topvert(0.0);
		horizontal_bar(0.0, 2.0 * npart * Rad, 0.0, RIGHT_BALL);
		advance(npart, 1.5);
		break;
    case 'M':	topvert(0.0);
		topvert(4.0 * mpart * Rad);
		cylinder(2.0 * mpart * Rad, 2.0 * Rad, 0.0, TOP, BOT_BALL);
		cylinder(2.0 * mpart * Rad, 2.0 * Rad, 4.0 * mpart * Rad, TOP, NO_BALLS);
		advance(2.0 * mpart, 1.5);
		break;
    case 'N':	topvert(0.0);
		topvert(2.0 * npart * Rad);
		cylinder(2.0 * npart * Rad, 0.0, 0.0, TOP, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'Q':	cylinder(2.0*Rad, 0.0, 1.5*Rad, Rad/2.0, BOTH_BALLS);
	        /* FALLTHROUGH */
    case 'O':	egg_toroid();
		advance(1.0, 1.5);
		break;
    case 'R':	/* based on B */
		cylinder(npart * Rad + TOP / 4.0, 0.0, Rad * npart , TOP / 2.0, BOT_BALL);
	        /* FALLTHROUGH */
    case 'P':	/* based on B */
	        vertical_bar(0.0, 0.0, TOP, BOTH_BALLS);
		horizontal_bar(0.0, npart * Rad, TOP, NO_BALLS);
		horizontal_bar(0.0, npart * Rad, TOP/2.0, NO_BALLS);
		toroid(npart * Rad, 3.0*TOP/4.0, (TOP/4.0)/Rad, -90.0, 90.0, NO_BALLS);
		advance(npart/2.0 + TOP/8.0/Rad, 1.5);
		break;
    case '$':  { double x, r, cwi, l;
		  r = 0.35;	/* factor for small radius */
		  l = (1.0 - r) * TOP/2.0;
		  cwi = RAD(sangle);
		  x = cos(cwi) * l + r * TOP/2.0;
                 vertical_bar(2.0*x/3.0, -rad, TOP+rad, BOTH_BALLS);
		 vertical_bar(4.0*x/3.0, -rad, TOP+rad, BOTH_BALLS);
               }
	        /* FALLTHROUGH */
    case 'S':	{   double x, r, cwi, beta, l;
		    r = 0.40;	/* was 0.35 factor for small radius */
		    cwi = RAD(sangle);
		    l = (1.0 - r) * TOP/2.0;
		    x = cos(cwi) * l + r * TOP/2.0;
		    toroid(x, TOP/2.0, TOP/2.0/Rad, sangle, 180.0-sangle, FIRST_BALL);
		    toroid(x, TOP/2.0, TOP/2.0/Rad, 180.0+sangle, -sangle, FIRST_BALL);
		    beta = asin(r / (1.0 - r));	/* cwi - beta is the slope of center line segment */
		    if (beta <= 0.0 || beta >= cwi) {
			fprintf(stderr, "internal parameters incorrect\n");
			exit(1);
		    }
		    toroid(x - l * cos(cwi), TOP/2.0 + l * sin(cwi),
			TOP/2.0 * r/Rad, 180.0 - sangle, 270.0 - sangle + DEGREE(beta), NO_BALLS);
		    toroid(x + l * cos(cwi), TOP/2.0 - l * sin(cwi),
			TOP/2.0 * r/Rad, -sangle, 90.0 - sangle + DEGREE(beta), NO_BALLS);
		    l = sqrt(1.0 - 2.0 * r) * TOP/2.0;
		    cylinder(x + l * cos(cwi-beta), TOP/2.0 - l * sin(cwi-beta),
			     x - l * cos(cwi-beta), TOP/2.0 + l * sin(cwi-beta), NO_BALLS);
		    advance(x / Rad, 1.5);
		}
		break;
    case 'T':	topvert(Rad);
		horizontal_bar(0.0, 2.0 * Rad, TOP, BOTH_BALLS);
		advance(1.0, 1.5);
		break;
    case 'U':	toroid(npart * Rad, npart * Rad, npart, 180.0, 0.0, NO_BALLS);
		topvert(2.0 * npart * Rad); 
		vertical_bar(0.0, npart * Rad, TOP, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'V':	cylinder(npart * Rad, 0.0, 0.0, TOP, BOTH_BALLS);
		cylinder(npart * Rad, 0.0, 2.0 * npart * Rad, TOP, TOP_BALL);
		advance(npart, 1.5);
		break;
    case 'W':	cylinder(      mpart * Rad, 0.0, 0.0              , TOP, BOTH_BALLS);
		cylinder(      mpart * Rad, 0.0, 2.0 * mpart * Rad, TOP, NO_BALLS);
		cylinder(3.0 * mpart * Rad, 0.0, 2.0 * mpart * Rad, TOP, BOTH_BALLS);
		cylinder(3.0 * mpart * Rad, 0.0, 4.0 * mpart * Rad, TOP, TOP_BALL);
		advance(2.0 * mpart, 1.5);
		break;
    case 'X':	cylinder(0.0, 0.0, 2.0 * npart * Rad, TOP, BOTH_BALLS);
		cylinder(2.0 * npart * Rad, 0.0, 0.0, TOP, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'Y':  vertical_bar(npart * Rad, 0.0, 1.5 * Rad, BOT_BALL);
		cylinder(npart * Rad, 1.5 * Rad, 0.0, TOP, TOP_BALL);
		cylinder(npart * Rad, 1.5 * Rad, 2.0 * npart * Rad, TOP, TOP_BALL);
		advance(npart, 1.5);
		break;
    case 'Z':	horizontal_bar(0.0, 2.0 * npart * Rad, 0.0, BOTH_BALLS);
		horizontal_bar(0.0, 2.0 * npart * Rad, TOP, BOTH_BALLS);
		cylinder(0.0, 0.0, 2.0 * npart * Rad, TOP, NO_BALLS);
		advance(npart, 1.5);
		break;
    case '':	cylinder(0.0, 0.0, Rad, TOP, BOTH_BALLS);
		cylinder(2.0 * Rad, 0.0, Rad, TOP, BOT_BALL);
		horizontal_bar(abar * Rad, (2.0 - abar) * Rad, TOP*abar, NO_BALLS);
		ball(0.0       + umlautdx, 2.0 * Rad + topsize, TRUE);
		ball(2.0 * Rad - umlautdx, 2.0 * Rad + topsize, TRUE);
		advance(1.0, 1.5);
		break;
    case '':	egg_toroid();
		ball(0.0               + umlautdx, 2.0 * Rad + topsize, TRUE);
		ball(2.0 * npart * Rad - umlautdx, 2.0 * Rad + topsize, TRUE);
		advance(1.0, 1.5);
		break;
    case '':	toroid(npart * Rad, npart * Rad, npart, 180.0, 0.0, NO_BALLS);
		topvert(2.0 * npart * Rad); 
		vertical_bar(0.0, npart * Rad, TOP, BOTH_BALLS);
		ball(0.0               + umlautdx, 2.5 * Rad + topsize, TRUE);
		ball(2.0 * npart * Rad - umlautdx, 2.5 * Rad + topsize, TRUE);
		advance(npart, 1.5);
		break;


    case 'a':	std_toroid();
		smallvert(2.0 * Rad);
		advance(1.0, 1.5);
		break;
    case 'b':	topvert(0.0);
		std_toroid();
		advance(1.0, 1.5);
		break;
    case 'c':	toroid(Rad, Rad, 1.0, cangle, -cangle, BOTH_BALLS);
		advance(1.0, 1.0);
		break;
    case 'd':	std_toroid();
		topvert(2.0 * Rad);
		advance(1.0, 1.5);
		break;
    case 'e':	toroid(Rad, Rad, 1.0, 0.0, -eangle, BOTH_BALLS);
		horizontal_bar(0.0, 2.0 * Rad, Rad, NO_BALLS);
		advance(1.0, 1.5);
		break;
    case 'f':	vertical_bar(2.0 * rad, -botsize, TOP - fpart * Rad, BOTH_BALLS);
		toroid(fpart * Rad + 2.0 * rad, TOP - fpart * Rad, fpart, sangle, 180.0, FIRST_BALL);
		horizontal_bar(0.0, 4.0 * rad, 2.0 * Rad, BOTH_BALLS);
		advance(fpart * (1.0 + cos(M_PI * (double)sangle / 180.0)) / 2.0, 2.5);
		break;
    case 'g':	std_toroid();
		toroid(Rad, Rad-botsize, 1.0, 180.0+sangle, 0.0, FIRST_BALL);
		vertical_bar(2.0 * Rad, -botsize + Rad, 2.0 * Rad, BOTH_BALLS);
		advance(1.0, 1.5);
		break;
    case 'h':	toroid(npart * Rad, (2.0 - npart) * Rad, npart, 0.0, 180.0, NO_BALLS);
		topvert(0.0);
		vertical_bar(2.0 * Rad * npart, 0.0, (2.0 - npart) * Rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'i':  ball(0.0, 2.0 * Rad + topsize/2.0, TRUE);
		smallvert(0.0);
		advance(0.0, 1.5);
		break;
    case 'j':  
		advance(fpart/2.0 * (1.0 + cos(RAD(sangle))), 0.0);
                ball(0.0, 2.0 * Rad + topsize/2.0, TRUE);
		vertical_bar(0.0, -topsize + fpart * Rad, 2.0 * Rad, BOTH_BALLS);
		toroid(-fpart * Rad, -topsize + fpart * Rad, fpart, 180.0 + sangle, 0.0, FIRST_BALL);
		advance(0.0, 1.5);
		break;
    case 'k':	topvert(0.0);
		cylinder(0.0, 2.0/3.0 * Rad, kwidth * Rad, 2.0 * Rad, TOP_BALL);
		cylinder(kwidth * Rad, 0.0, Rad * kwidth/4.0, Rad, BOT_BALL);
		advance(kwidth / 2.0, 1.5);
		break;
    case 'l':	/* t without bar */
		vertical_bar(0.0, fpart * Rad, TOP, BOTH_BALLS);
		toroid(fpart * Rad, fpart * Rad, fpart, 180.0, -sangle, END_BALL);
		advance(fpart * (1.0 + cos(M_PI * (double)sangle / 180.0)) / 2.0, 1.5);
		break;
    case 'm':	toroid(mpart *       Rad, (2.0-mpart) * Rad, mpart, 0.0, 180.0, NO_BALLS);
		toroid(mpart * 3.0 * Rad, (2.0-mpart) * Rad, mpart, 0.0, 180.0, NO_BALLS);
		smallvert(0.0);
		vertical_bar(2.0 * Rad * mpart, 0.0, (2.0 - mpart) * Rad, BOTH_BALLS);
		vertical_bar(4.0 * Rad * mpart, 0.0, (2.0 - mpart) * Rad, BOTH_BALLS);
		advance(2.0 * mpart, 1.5);
		break;
    case 'n':	toroid(npart * Rad, (2.0 - npart) * Rad, npart, 0.0, 180.0, NO_BALLS);
		smallvert(0.0);
		vertical_bar(2.0 * Rad * npart, 0.0, (2.0 - npart) * Rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'o':	std_toroid();
		advance(1.0, 1.5);
		break;
    case 'p':	std_toroid();
		botvert(0.0);
		advance(1.0, 1.5);
		break;
    case 'q':	std_toroid();
		botvert(2.0 * Rad);
		advance(1.0, 1.5);
		break;
    case 'r':	/* toroid is very much like an f */
		smallvert(0.0);
		toroid(fpart * Rad, 2.0 * Rad - fpart * Rad, fpart, sangle, 180.0, FIRST_BALL);
		advance(fpart, 1.5);
		break;
    case 's':	{   double x, r, cwi, beta, l;
		    r = 0.35;	/* factor for small radius */
		    cwi = M_PI / 180.0 * (double)sangle;
		    l = (1.0 - r) * Rad;
		    x = cos(cwi) * l + r * Rad;
		    toroid(x, Rad, 1.0, sangle, 180.0-sangle, BOTH_BALLS);
		    toroid(x, Rad, 1.0, 180.0+sangle, -sangle, BOTH_BALLS);
		    beta = asin(r / (1.0 - r));	/* cwi - beta is the slope of center line segment */
		    if (beta <= 0.0 || beta >= cwi) {
			fprintf(stderr, "internal parameters incorrect\n");
			exit(1);
		    }
		    toroid(x - l * cos(cwi), Rad + l * sin(cwi),
			r, 180.0 - sangle, 270.0 - sangle + DEGREE(beta), NO_BALLS);
		    toroid(x + l * cos(cwi), Rad - l * sin(cwi),
			r, -sangle, 90.0 - sangle + DEGREE(beta), NO_BALLS);
		    l = sqrt(1.0 - 2.0 * r) * Rad;
		    cylinder(x + l * cos(cwi-beta), Rad - l * sin(cwi-beta),
			     x - l * cos(cwi-beta), Rad + l * sin(cwi-beta), NO_BALLS);
		    advance(x / Rad, 1.5);
		}
		break;
    case 't':	/* see f */
		vertical_bar(2.0 * rad, fpart * Rad, TOP, BOTH_BALLS);
		toroid(fpart * Rad + 2.0 * rad, fpart * Rad, fpart, 180.0, -sangle, END_BALL);
		horizontal_bar(0.0, 4.0 * rad, 2.0 * Rad, BOTH_BALLS);
		advance(fpart * (1.0 + cos(M_PI * (double)sangle / 180.0)) / 2.0, 2.5);
		break;
    case 'u':	toroid(npart * Rad, npart * Rad, npart, 180.0, 0.0, NO_BALLS);
		smallvert(2.0 * npart * Rad);
		vertical_bar(0.0, npart * Rad, 2.0 * Rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'v':	cylinder(npart * Rad, 0.0, 0.0, 2.0 * Rad, BOTH_BALLS);
		cylinder(npart * Rad, 0.0, 2.0 * npart * Rad, 2.0 * Rad, TOP_BALL);
		advance(npart, 1.5);
		break;
    case 'w':	cylinder(      mpart * Rad, 0.0, 0.0              , 2.0 * Rad, BOTH_BALLS);
		cylinder(      mpart * Rad, 0.0, 2.0 * mpart * Rad, 2.0 * Rad, NO_BALLS);
		cylinder(3.0 * mpart * Rad, 0.0, 2.0 * mpart * Rad, 2.0 * Rad, BOTH_BALLS);
		cylinder(3.0 * mpart * Rad, 0.0, 4.0 * mpart * Rad, 2.0 * Rad, TOP_BALL);
		advance(2.0 * mpart, 1.5);
		break;
    case 'x':	cylinder(0.0, 0.0, 2.0 * npart * Rad, 2.0 * Rad, BOTH_BALLS);
		cylinder(2.0 * npart * Rad, 0.0, 0.0, 2.0 * Rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'y':	/* y: a v with longer cylinder */
		cylinder(npart * Rad, 0.0, 0.0, 2.0 * Rad, TOP_BALL);
		cylinder(npart*(Rad-botsize/2.0), -botsize, 2.0 * npart * Rad, 2.0 * Rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
    case 'z':	horizontal_bar(0.0, 2.0 * npart * Rad, 0.0, BOTH_BALLS);
		horizontal_bar(0.0, 2.0 * npart * Rad, 2.0 * Rad, BOTH_BALLS);
		cylinder(0.0, 0.0, 2.0 * npart * Rad, 2.0 * Rad, NO_BALLS);
		advance(npart, 1.5);
		break;
    case '':  std_toroid();   /* Umlaut ae */
		smallvert(2.0 * Rad);
		ball(0.0       + umlautdx, 2.0 * Rad + topsize/2.0, TRUE);
		ball(2.0 * Rad - umlautdx, 2.0 * Rad + topsize/2.0, TRUE);
		advance(1.0, 1.5);
		break;
    case '':	std_toroid();
		ball(0.0       + umlautdx, 2.0 * Rad + topsize/2.0, TRUE);
		ball(2.0 * Rad - umlautdx, 2.0 * Rad + topsize/2.0, TRUE);
		advance(1.0, 1.5);
		break;
    case '':	toroid(npart * Rad, npart * Rad, npart, 180.0, 0.0, NO_BALLS);
		smallvert(2.0 * npart * Rad);
		vertical_bar(0.0, npart * Rad, 2.0 * Rad, BOTH_BALLS);
		ball(0.0               + umlautdx, 2.0 * Rad + topsize/2.0, TRUE);
		ball(2.0 * npart * Rad - umlautdx, 2.0 * Rad + topsize/2.0, TRUE);
		advance(npart, 1.5);
		break;
    case '':	/* based on f */
               {
               double szbend = RAD(0.0);
               double szang1 = RAD(280.0);
	        double szang2 = RAD(-110.0);
	        double szrad2 = (TOP - Rad * szpart * (1.0 - sin(szang1)))/
			        (1.0 + sin(RAD(180.0) + szbend + szang1))/Rad;
		advance(0.0, -1.0);
	        vertical_bar(2.0 * rad, -botsize, TOP - szpart * Rad, FIRST_BALL);
		toroid(szpart * Rad + 2.0 * rad, 
                    TOP - szpart * Rad, szpart, DEGREE(szang1), 180.0, NO_BALLS);
               toroid(szpart * Rad * (1.0 + cos(szang1)) - szrad2 * cos(RAD(180.0) + szbend + szang1),
                     TOP + szpart * Rad * (-1.0 + sin(szang1)) - szrad2 * Rad * sin(RAD(180.0) + szbend + szang1),
                     szrad2, DEGREE(szang2), DEGREE(RAD(180.0) + szbend + szang1), BOTH_BALLS);
		advance((szpart*(1.0+cos(szang1)) + szrad2*(1.0-cos(RAD(180.0) + szbend + szang1)))/2.0, 2.5);
		break;
               }
    case '0':  egg_toroid();
	       cylinder(0.0, 0.0, 2.0 * Rad, TOP, BOTH_BALLS);
	       advance(1.0, 1.5);
	       break;
    case '1':  topvert(Rad/2.0);
               cylinder(0.0, TOP - Rad/2.0, Rad/2.0, TOP, BOTH_BALLS);
	       advance(1.0/3.0, 1.5);
	       break;
    case '2':  {double x,y;
		double angle2;
#define rad2	(0.9*Rad)
		angle2 = -asin(2*rad2/(TOP-rad2));
		x = rad2       + rad2*cos(angle2);
		y = (TOP-rad2) + rad2*sin(angle2);
                toroid(rad2,TOP-rad2,rad2/Rad,DEGREE(angle2),180,END_BALL);
		cylinder(0.0, 0.0, x, y, NO_BALLS);
	        horizontal_bar(0.0, 2.0 * rad2, 0.0, BOTH_BALLS);
		advance(1.0, 1.5);
	       }
	       break;
#define angle3	50
    case '3':   toroid(TOP/4.0* (sin(RAD(angle3))),
		       3.0*TOP/4.0,TOP/4.0/Rad,270,90+angle3,BOTH_BALLS);
	        toroid(TOP/4.0* (sin(RAD(angle3))),    
		       TOP/4.0,TOP/4.0/Rad,270-angle3,90,BOT_BALL);
		advance(TOP/8.0*(1.0 + sin(RAD(angle3)))/Rad, 1.5);
		break;
    case '4':	topvert(1.5 * npart * Rad);
		cylinder(0.0, Rad, 1.5 * npart * Rad, TOP, NO_BALLS);
	        horizontal_bar(0.0, 2.0 * npart * Rad, Rad, BOTH_BALLS);
		advance(npart, 1.5);
		break;
#define angle5	130.0
    case '5':	{double x,y;
		 x = Rad + Rad * cos(RAD(angle5));
		 y = Rad + Rad * sin(RAD(angle5));
	         horizontal_bar(x, 2.0 * Rad, TOP, BOTH_BALLS);
		 vertical_bar(x, y, TOP, NO_BALLS);
		 toroid(Rad, Rad, 1.0, 205.0, angle5, BOTH_BALLS);
		 advance(1.0, 1.5);
	        }
		break;
    case '6':	std_toroid();
		vertical_bar(0.0, Rad, TOP-Rad, NO_BALLS);
	        toroid(Rad, TOP - Rad, 1.0, 45, 180, FIRST_BALL);
		advance(1.0, 1.5);
		break;
    case '7':	horizontal_bar(0.0, 2.0 * Rad, TOP, BOTH_BALLS);
		cylinder(0.0, 0.0, 2.0 * Rad, TOP, BOT_BALL);
		advance(1.0, 1.5);
		break;
    case '8':   toroid(TOP/4.0, 3.0*TOP/4.0, TOP/4.0/Rad,0,0,NO_BALLS);
	        toroid(TOP/4.0,     TOP/4.0, TOP/4.0/Rad,0,0,NO_BALLS);
		advance(TOP/4.0/Rad, 1.5);
		break;
    case '9':	std_toroid2(0.0, TOP- 2.0 * Rad);
		toroid(Rad, Rad, 1.0, 225.0, 0.0, FIRST_BALL);
		vertical_bar(2.0*Rad, Rad, TOP - Rad, NO_BALLS);
		advance(1.0, 1.5);
		break;

    case '~':
    case '{':
    case '}':
    case '&':
    case  0: case  1:case  2:case  3:case  4:case  5:case  6:case  7:case  8:case  9:
    case 10: case 11:case 12:case 13:case 14:case 15:case 16:case 17:case 18:case 19:
    case 20: case 21:case 22:case 23:case 24:case 25:case 26:case 27:case 28:case 29:
    case 30: case 31:
    default:
    	        fprintf(stderr, "bad character: %c\n", c);
    case ' ':	advance(0.5, 1.5);
	        break;
    }
}

int main(int argc, char *argv[])
{   char *s;
    if (argc > 4) {
	fprintf(stderr, "usage: povwrite (text) object_name [outfile]\n");
	exit(1);
    }
    if (argc == 4) /* reuse standard output file descriptor */
	if (!freopen(argv[3], "w", stdout)) {
	    fprintf(stderr, "cannot open file %s\n", argv[3]);
	    exit(2);
	}

    printf("// POV-Ray include file generated by string generator povwrite\n");
    printf("// this file defines the string\n// %s\n",argv[1]);
    printf("#version 2.0\n");
    declare_toroid();
    uv_toroid();
    printf("#declare %s = object {\n    union {\n",argv[2] ? argv[2] :"Default");
    for (s = argv[1]; *s; ++s)
	writeletter(*s);
    if (!dy && !dz) {
	printf("\t}\n    bounded_by {\n\tintersection {\n");
	printf("\t    plane { <0, 0,  1>, %f }\n\t    plane { <0, 0, -1>, %f }\n",
	    zmax + 1.001 * rad, -zmin + 1.001 * rad);
	printf("\t    plane { <0,  1, 0>, %f }\n\t    plane { <0, -1, 0>, %f }\n",
	    ymax + 1.001 * rad, -ymin + 1.001 * rad);
	printf("\t    plane { < 1, 0, 0>, %f }\n\t    plane { <-1, 0, 0>, %f }\n",
	    xmax + 1.001 * rad, -xmin + 1.001 * rad);
	printf("\t}\n    }\n");
    }
    printf("}\n");	/* end object */
    return 0;
}
